import duckdb
from core_pro.ultilities import make_sync_folder
from pathlib import Path
import sys
sys.path.extend([str(Path.home() / 'PycharmProjects/item_matching')])

from src.item_matching.func.function_image import PipelineImage
from src.item_matching.func.function_text import PipelineText
from src.item_matching.main_match import PipelineMatch, MatchInput
from src.item_matching.pipeline.rerank import ReRank, ReRankConfig


# path
cluster = 'FMCG'
path = make_sync_folder('Item_Matching_Test')

file = path / f'data_sample_{cluster}_0.parquet'
file_inner = path / 'inner.parquet'

# data loading
query = f"""
with base as (
select * exclude(level1_global_be_category, level2_global_be_category, level3_global_be_category)
, level1_global_be_category || '__' || level2_global_be_category || '__' || level3_global_be_category category
from read_parquet('{file}')
)
select *
from base
"""
df = (
    duckdb.sql(query).pl()
    .unique(['item_id'])
)

# text
df = PipelineText().run(df)
df.write_parquet(file_inner)

config = MatchInput(
    ROOT_PATH=path,
    PATH_INNER=file_inner,
    INNER=True ,
    MATCH_BY='text',
    COL_CATEGORY='category',
    TOP_K=10,
    EXPLODE=False,
)
PipelineMatch(config).run()

# image
df, df_img = PipelineImage(path / f'download_img_{cluster}').run(df)
df.write_parquet(file_inner)

config = MatchInput(
    ROOT_PATH=path,
    PATH_INNER=file_inner,
    INNER=True ,
    MATCH_BY='image',
    COL_CATEGORY='category',
    TOP_K=10,
    EXPLODE=False,
)
PipelineMatch(config).run()

# rerank
config = ReRankConfig(ROOTPATH=path, EXPLODE=True)
ReRank(config).run()



import polars as pl
import duckdb
from pathlib import Path
from pydantic import BaseModel, Field, computed_field
from tqdm import tqdm
from rich import print
from core_pro.ultilities import make_dir
from ..func.post_processing import PostProcessing, data_explode_list


class ReRankConfig(BaseModel):
    ROOTPATH: Path = Field(default=None)
    EXPLODE: bool = Field(default=False)

    @computed_field
    @property
    def path_result(self) -> Path:
        path_result = self.ROOTPATH / 'result_match_rerank'
        make_dir(path_result)
        return path_result

    @computed_field
    @property
    def path_text(self) -> Path:
        return self.ROOTPATH / 'result_match_text'

    @computed_field
    @property
    def path_image(self) -> Path:
        return self.ROOTPATH / 'result_match_image'

    @computed_field
    @property
    def file_text(self) -> list:
        return [*self.path_text.glob('*')]

    @computed_field
    @property
    def file_image(self) -> list:
        return [*self.path_image.glob('*')]


class ReRank:
    def __init__(self, record: ReRankConfig):
        # all category
        self.all_category = set([i.stem for i in record.file_text] + [i.stem for i in record.file_image])

        # config
        self.EXPLODE = record.EXPLODE

        # path
        self.path_result = record.path_result
        self.path_image = record.path_image
        self.path_text = record.path_text

        # file
        self.file_text = record.file_text
        self.file_image = record.file_image

    def _data_check(self, category: str):
        paths = {'text': self.path_text, 'image': self.path_image}
        df_dict, total_cols = {}, []
        for name, path in paths.items():
            df = pl.read_parquet(path / f'{category}.parquet')
            if self.EXPLODE:
                df = data_explode_list(df)
            total_cols.extend(df.columns)
            df_dict[name] = df
        total_cols = set(total_cols)

        # select cols
        patterns = ['item_name_clean', 'file_path', 'rnk', 'exists', 'score_text_embed', 'score_image_embed']
        select_cols = [col for col in total_cols if not any(pattern in col for pattern in patterns)]
        duckdb_cols = '\n, '.join([f'{i}' for i in select_cols])
        return df_dict, duckdb_cols

    def rerank_score(self, data_text, data_image, duckdb_cols):
        # query
        query = f"""
            -- combine text & image
            with base as (
                select {duckdb_cols}
                , 'score_text' match_type
                , score_text_embed score
                from data_text
                union all
                select {duckdb_cols}
                , 'score_image' match_type
                , score_image_embed score
                from data_image
            ) 
            -- pivot scores into 2 cols
            , pivot_tab as (
                PIVOT base
                ON match_type
                USING sum(score)
                GROUP BY {duckdb_cols}
            )
            -- calculate mean, max rerank
            , cal_tab as (
                select *
                , (coalesce(score_text, 0) + coalesce(score_image, 0)) / 2 as score_mean
--                 , greatest(score_text, score_image) as score_max
                from pivot_tab
                where db_item_id != q_item_id
            )
            -- rank matches
            select * 
            , row_number() OVER (PARTITION BY q_item_id ORDER BY score_mean desc) ranking
            from cal_tab
            """
        return duckdb.sql(query).pl()

    def run(self):
        total_cat = len(self.all_category)
        for cat in tqdm(self.all_category, total=total_cat, desc='ReRanking'):
            df_dict, duckdb_cols = self._data_check(cat)
            df = self.rerank_score(data_text=df_dict['text'], data_image=df_dict['image'], duckdb_cols=duckdb_cols)
            df = PostProcessing(df).run()
            df.write_parquet(self.path_result / f'{cat}.parquet')
        print(f'[RERANK]: Done {total_cat} categories')



from pydantic import BaseModel, computed_field, Field
from typing import List, Any
from pathlib import Path
import duckdb
from rich import print


required_columns = {
    'item_name',
    'level1_global_be_category',
    'level2_global_be_category',
    'level3_global_be_category'
}

class ValidationUserData(BaseModel):
    MATCH_BY: str = Field(default='text')
    FILE: Path
    MODE: str = Field(default='')
    INPUT_COLUMNS: List[str] = Field(default=[])

    @computed_field
    @property
    def check_file_extension(self) -> str:
        if self.FILE.suffix not in ['.csv', '.parquet']:
            return '[FILE TYPE]: Please specify a .csv or .parquet file'

    @computed_field
    @property
    def required_columns(self) -> set:
        if self.MATCH_BY == 'image':
            required_columns.add('image_url')
        return required_columns

    @computed_field
    @property
    def read_data(self) -> str:
        try:
            data = duckdb.sql(f"select * from read_{self.FILE.suffix[1:]}('{self.FILE}') limit 10").pl()
            self.INPUT_COLUMNS = data.columns
        except Exception as errors:
            return f"[UNKNOWN READING] Your **{self.MODE}** file may be damaged: {errors}"

    @computed_field
    @property
    def validation(self) -> str:
        INPUT = set(self.INPUT_COLUMNS)
        missing_columns = list(self.required_columns - INPUT)
        if missing_columns:
            message = (
                f"[REQUIRED COLUMNS]: Missing {missing_columns}. "
                f"Your columns: {self.INPUT_COLUMNS}"
            )
            return message

    @computed_field
    @property
    def summary(self) -> list:
        pipe = [
            self.read_data,
            self.validation,
            self.check_file_extension,
        ]
        return [i for i in pipe if i]


# valid = UserData(
#     INPUT_COLUMNS=['item_name', 'level1_global_be_category', 'level2_global_be_category', 'level3_global_be_category'],
#     MATCH_BY='text',
#     FILE= Path('file.parquet'),
#     MODE='db'
# ).summary
# print(valid)
#
# valid = UserData(
#     INPUT_COLUMNS=['item_name', 'level1_global_be_category'],
#     MATCH_BY='image',
#     FILE= Path('file.parquet'),
#     MODE='db'
# ).summary
# print(valid)



import os, subprocess
from pathlib import Path
from rich import print
import duckdb
import polars as pl
from core_pro import DataPipeLine
from core_pro.ultilities import make_sync_folder, make_dir
from core_eda import EDA_Dataframe


# download parquet
path = make_sync_folder('Item_Matching_Test')
path_sql = Path.home() / 'PycharmProjects/item_matching/notebooks/data/data.sql'

for c in ['ELHA', 'Lifestyle', 'Fashion', 'FMCG']:
    file = path / f'data_sample_{c}.parquet'
    if not file.exists():
        print(f'Download {file.stem}')
        sql = open(str(path_sql)).read().format(c)
        df = DataPipeLine(sql).run_presto_to_df(save_path=file)
    else:
        df = pl.read_parquet(file)
        print(f'File {file.stem} exists')

    # download images
    query = f"""
    select *
    ,concat('http://f.shopee.vn/file/', UNNEST(array_slice(string_split(images, ','), 1, 1))) image_url
    from parquet_scan('{str(path / f'{file.stem}.parquet')}')
    order by item_id, images
    """
    df = (
        duckdb.sql(query).pl()
        .unique(['item_id'])
    )
    EDA_Dataframe(df, ['item_id']).check_duplicate()

    path_parq = path / f'{file.stem}_0.parquet'
    df.write_parquet(path_parq)

    path_image = path / f'download_img_{c}'
    make_dir(path_image)
    folder = path_image / f'img_'
    if not folder.exists():
        os.chdir(str(path_image))
        command = (
            f"img2dataset --url_list={str(path_parq)} "
            f"--output_folder=img_/ "
            f"--processes_count=16 "
            f"--thread_count=32 "
            f"--image_size=224 "
            f"--output_format=files "
            f"--input_format=parquet "
            f"--url_col=image_url "
            f"--number_sample_per_shard=50000 "
        )
        subprocess.run(command, shell=True)



from pathlib import Path
import polars as pl
from time import perf_counter
from autofaiss import build_index
from datasets import concatenate_datasets, load_from_disk
from pydantic import BaseModel, Field, computed_field
import numpy as np
from rich import print
from core_pro.ultilities import create_batch_index, make_dir
from ..func.post_processing import data_explode_list


class ConfigQuery(BaseModel):
    ROOT_PATH: Path = Field(default=None)
    QUERY_SIZE: int = Field(default=50_000)
    MATCH_BY: str = Field(default='text')
    TOP_K: int = Field(default=10)

    @computed_field
    @property
    def col_embedding(self) -> str:
        dict_ = {'image': 'image_embed'}
        return dict_.get(self.MATCH_BY, 'text_embed')

    @computed_field
    @property
    def path_array_db(self) -> Path:
        return self.ROOT_PATH / f'db_array'

    @computed_field
    @property
    def path_ds_db(self) -> Path:
        return self.ROOT_PATH / f'db_ds'

    @computed_field
    @property
    def path_ds_q(self) -> Path:
        return self.ROOT_PATH / f'q_ds'

    @computed_field
    @property
    def path_ds_inner(self) -> Path:
        return self.ROOT_PATH / f'ds'

    @computed_field
    @property
    def path_array_inner(self) -> Path:
        return self.ROOT_PATH / f'array'

    @computed_field
    @property
    def path_index(self) -> Path:
        return self.ROOT_PATH / f'index'

    @computed_field
    @property
    def path_result_query_score(self) -> Path:
        path_result = self.ROOT_PATH / f'result'
        make_dir(path_result)
        return path_result

    @computed_field
    @property
    def path_result_final(self) -> Path:
        path_result = self.ROOT_PATH / f'result_match_{self.MATCH_BY}'
        make_dir(path_result)
        return path_result


class BuildIndexAndQuery:
    def __init__(
            self,
            file_export_name: str,
            config: ConfigQuery,
            inner: bool = False,
            explode: bool = True
    ):
        self.TOP_K = config.TOP_K
        self.QUERY_SIZE = config.QUERY_SIZE
        self.col_embedding = config.col_embedding
        self.inner = inner
        self.explode = explode
        self.sort_key_ds = lambda x: int(x.stem)
        self.sort_key_result = lambda x: int(x.stem.split('_')[1])

        # index
        self.path_index = config.path_index
        self.file_index = self.path_index / f'ip.index'
        self.file_index_json = str(self.path_index / f'index.json')

        # array
        self.path_array_db = config.path_array_inner if inner else config.path_array_db

        # ds
        self.dataset_dict = {
            'db_ds_path': config.path_ds_db,
            'q_ds_path': config.path_ds_q,
            'inner_ds_path': config.path_ds_inner,
        }

        # result
        self.path_result_query_score = config.path_result_query_score
        self.path_result_final = config.path_result_final
        self.file_export_final = self.path_result_final / f'{file_export_name}.parquet'

    def build(self):
        # Build index
        start = perf_counter()
        if not self.file_index.exists():
            print(f'[BuildIndex] Start')
            build_index(
                str(self.path_array_db),
                index_path=str(self.file_index),
                index_infos_path=self.file_index_json,
                save_on_disk=True,
                metric_type='ip',
                verbose=30,
            )
            print(f'[BuildIndex] Time finished: {perf_counter() - start:,.2f}s')
        else:
            print(f'[BuildIndex] Index is existed')

    def load_dataset(self):
        dataset = {}
        if self.inner:
            for i in ['db', 'q']:
                dataset[i] = concatenate_datasets([
                    load_from_disk(str(f))
                    for f in sorted(self.dataset_dict[f'inner_ds_path'].glob('*'), key=self.sort_key_ds)
                ])

                for c in dataset[i].column_names:
                    if c != self.col_embedding:
                        dataset[i] = dataset[i].rename_column(c, f'{i}_{c}')

        else:
            for i in ['db', 'q']:
                dataset[i] = concatenate_datasets([
                    load_from_disk(str(f))
                    for f in sorted(self.dataset_dict[f'{i}_ds_path'].glob('*'), key=self.sort_key_ds)
                ])

        # Add index
        dataset['db'].load_faiss_index(self.col_embedding, self.file_index)
        return dataset['db'], dataset['q']

    def query(self):
        # Load dataset
        dataset_db, dataset_q = self.load_dataset()
        if len(dataset_db) < 2 or len(dataset_q) < 2:
            print(f'[BuildIndex] DB/Q dataset < 2 rows')
            return None

        # Batch query
        run = create_batch_index(len(dataset_q), self.QUERY_SIZE)
        num_batches = len(run)
        for i, val in run.items():
            # init
            file_name_result = self.path_result_query_score / f'result_{i}.parquet'
            file_name_score = self.path_result_query_score / f'score_{i}.parquet'
            if file_name_result.exists():
                continue

            # query
            start_idx, end_idx = val[0], val[-1]
            start_batch = perf_counter()
            score, result = dataset_db.get_nearest_examples_batch(
                self.col_embedding,
                dataset_q[start_idx:end_idx][self.col_embedding],
                k=self.TOP_K,
            )
            # export
            for arr in result:
                del arr[self.col_embedding]  # prevent memory leaks
            df_result = pl.DataFrame(result)
            df_result.write_parquet(file_name_result)

            # track errors
            if df_result.shape[0] == 0:
                print(f"[red]No matches found for {i}[/]")
                continue

            dict_ = {f'score_{self.col_embedding}': [list(np.round(arr, 6)) for arr in score]}
            df_score = pl.DataFrame(dict_)
            df_score.write_parquet(file_name_score)

            # log
            end = perf_counter() - start_batch
            print(f"[Query] Batch {i}/{num_batches - 1} match result shape: {df_result.shape} {end:,.2f}s")
            del score, result, df_score, df_result

        # Post process
        dataset_q = dataset_q.remove_columns(self.col_embedding)  # prevent polars issues
        del dataset_db

        df_score = (
            pl.concat([
                pl.read_parquet(f)
                for f in sorted(self.path_result_query_score.glob('score*.parquet'), key=self.sort_key_result)])
        )
        df_result = (
            pl.concat([
                pl.read_parquet(f)
                for f in sorted(self.path_result_query_score.glob('result*.parquet'), key=self.sort_key_result)])
        )

        df_match = pl.concat([dataset_q.to_polars(), df_result, df_score], how='horizontal')
        if self.explode:
            df_match = data_explode_list(df_match)

        df_match.write_parquet(self.file_export_final)
